#undef DEBUG
#include "internal.h"

#include "al.h"
#include "alc.h"

#include <stdlib.h>

#define STREAMBUFS 3

static ALCdevice *aldev = NULL;
static ALCcontext *alctx = NULL;

typedef struct {
	char used;
	ALuint source;
	char stream;
	jfaudplaymode supposedly;
	jfaud_buffer buffers[STREAMBUFS];
	char buffersused[STREAMBUFS];
} source;

static source *sources = NULL;

static struct jfaud_init_params *params;

#if (DYNAMIC_OPENAL == 1)

#ifdef _WIN32
# define WIN32_LEAN_AND_MEAN
# include <windows.h>
# define DLLNAME "openal32.dll"
#else
# include <dlfcn.h>
# define DLLNAME "libopenal.so"
#endif

static struct {
#ifdef _WIN32
	HMODULE dll;
#else
	void *dll;
#endif

	ALboolean   (ALCAPIENTRY *alcIsExtensionPresent)(ALCdevice *device, ALubyte *extName);
	ALCdevice *  (ALCAPIENTRY *alcOpenDevice)(ALubyte *tokstr);
	ALCcontext * (ALCAPIENTRY *alcCreateContext)(ALCdevice *dev, ALint* attrlist);
	ALubyte*    (ALCAPIENTRY *alcGetString)(ALCdevice *deviceHandle, ALCenum token);
	ALCenum      (ALCAPIENTRY *alcGetError)(ALCdevice *dev);
	ALboolean   (ALCAPIENTRY *alcMakeContextCurrent)(ALCcontext *context);
	ALvoid       (ALCAPIENTRY *alcDestroyContext)(ALCcontext* context);
	void         (ALCAPIENTRY *alcCloseDevice)(ALCdevice *dev);

	ALubyte*     (ALAPIENTRY  *alGetString)(ALenum param);
	ALenum       (ALAPIENTRY  *alGetError)(ALvoid);
	void         (ALAPIENTRY  *alGenSources)(ALsizei n, ALuint* sources);
	void         (ALAPIENTRY  *alGenBuffers)(ALsizei n, ALuint* buffers);
	void         (ALAPIENTRY  *alDeleteSources)(ALsizei n, ALuint* sources);
	void         (ALAPIENTRY  *alDeleteBuffers)(ALsizei n, ALuint* buffers);
	void         (ALAPIENTRY  *alListener3f)(ALenum pname, ALfloat f1, ALfloat f2, ALfloat f3);
	void         (ALAPIENTRY  *alListenerfv)(ALenum pname, ALfloat* param);
	void         (ALAPIENTRY  *alSourcePlay)(ALuint sid);
	void         (ALAPIENTRY  *alSourcePause)(ALuint sid);
	void         (ALAPIENTRY  *alSourceStop)(ALuint sid);
	void         (ALAPIENTRY  *alSourceRewind)(ALuint sid);
	void         (ALAPIENTRY  *alSourcei)(ALuint sid, ALenum param, ALint value);
	void         (ALAPIENTRY  *alSourcef)(ALuint sid, ALenum param, ALfloat value);
	void         (ALAPIENTRY  *alSource3f)(ALuint sid, ALenum param, ALfloat f1, ALfloat f2, ALfloat f3);
	void         (ALAPIENTRY  *alGetSourcei)(ALuint sid, ALenum pname, ALint* value);
	void         (ALAPIENTRY  *alSourceQueueBuffers)(ALuint sid, ALsizei numEntries, ALuint *bids);
	void         (ALAPIENTRY  *alSourceUnqueueBuffers)(ALuint sid, ALsizei numEntries, ALuint *bids);
	void         (ALAPIENTRY  *alBufferData)(ALuint buffer, ALenum format, ALvoid* data, ALsizei size, ALsizei freq);
} libopenal = {NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

#define alcIsExtensionPresent libopenal.alcIsExtensionPresent
#define alcOpenDevice libopenal.alcOpenDevice
#define alcCreateContext libopenal.alcCreateContext
#define alcGetString libopenal.alcGetString
#define alcGetError libopenal.alcGetError
#define alcMakeContextCurrent libopenal.alcMakeContextCurrent
#define alcDestroyContext libopenal.alcDestroyContext
#define alcCloseDevice libopenal.alcCloseDevice
#define alGetString libopenal.alGetString
#define alGetError libopenal.alGetError
#define alGenSources libopenal.alGenSources
#define alGenBuffers libopenal.alGenBuffers
#define alDeleteSources libopenal.alDeleteSources
#define alDeleteBuffers libopenal.alDeleteBuffers
#define alListener3f libopenal.alListener3f
#define alListenerfv libopenal.alListenerfv
#define alSourcePlay libopenal.alSourcePlay
#define alSourcePause libopenal.alSourcePause
#define alSourceStop libopenal.alSourceStop
#define alSourceRewind libopenal.alSourceRewind
#define alSourcei libopenal.alSourcei
#define alSourcef libopenal.alSourcef
#define alSource3f libopenal.alSource3f
#define alGetSourcei libopenal.alGetSourcei
#define alSourceQueueBuffers libopenal.alSourceQueueBuffers
#define alSourceUnqueueBuffers libopenal.alSourceUnqueueBuffers
#define alBufferData libopenal.alBufferData

#define testifloaded() if (libopenal.dll == NULL) return jfauderr_notinited

#else
#define testifloaded() { }
#endif


jfauderr jfaudopenal_init(void)
{
	ALenum err;
	int i,j,er=0;
	void *sym;

	if (alctx) return jfauderr_alreadyinited;

	params = jfaud_getparams();

#if (DYNAMIC_OPENAL == 1)
#ifdef _WIN32
	libopenal.dll = LoadLibrary(DLLNAME);
	
#define gpa(d,x) sym = GetProcAddress(d,x); { \
		if (!sym) { \
			params->call_logmsg(jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
			er = 1; \
		} \
	}
#else
	libopenal.dll = dlopen(DLLNAME, RTLD_GLOBAL|RTLD_NOW);
	dlerror();

#define gpa(d,x) dlsym(d,x); { \
		const char *s; \
		if (s = dlerror()) { \
			params->call_logmsg(jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
			er = 1; \
		} \
	}
#endif

	if (!libopenal.dll) {
		params->call_logmsg("Failed to load " DLLNAME " -- sound disabled.");
		return jfauderr_notinited;
	}

	alcIsExtensionPresent = (void*)gpa(libopenal.dll,"alcIsExtensionPresent");
	alcOpenDevice = (void*)gpa(libopenal.dll,"alcOpenDevice");
	alcCreateContext = (void*)gpa(libopenal.dll,"alcCreateContext");
	alcGetString = (void*)gpa(libopenal.dll,"alcGetString");
	alcGetError = (void*)gpa(libopenal.dll,"alcGetError");
	alcMakeContextCurrent = (void*)gpa(libopenal.dll,"alcMakeContextCurrent");
	alcDestroyContext = (void*)gpa(libopenal.dll,"alcDestroyContext");
	alcCloseDevice = (void*)gpa(libopenal.dll,"alcCloseDevice");
	
	alGetString = (void*)gpa(libopenal.dll,"alGetString");
	alGetError = (void*)gpa(libopenal.dll,"alGetError");
	alGenSources = (void*)gpa(libopenal.dll,"alGenSources");
	alGenBuffers = (void*)gpa(libopenal.dll,"alGenBuffers");
	alDeleteSources = (void*)gpa(libopenal.dll,"alDeleteSources");
	alDeleteBuffers = (void*)gpa(libopenal.dll,"alDeleteBuffers");
	alListener3f = (void*)gpa(libopenal.dll,"alListener3f");
	alListenerfv = (void*)gpa(libopenal.dll,"alListenerfv");
	alSourcePlay = (void*)gpa(libopenal.dll,"alSourcePlay");
	alSourcePause = (void*)gpa(libopenal.dll,"alSourcePause");
	alSourceStop = (void*)gpa(libopenal.dll,"alSourceStop");
	alSourceRewind = (void*)gpa(libopenal.dll,"alSourceRewind");
	alSourcei = (void*)gpa(libopenal.dll,"alSourcei");
	alSourcef = (void*)gpa(libopenal.dll,"alSourcef");
	alSource3f = (void*)gpa(libopenal.dll,"alSource3f");
	alGetSourcei = (void*)gpa(libopenal.dll,"alGetSourcei");
	alSourceQueueBuffers = (void*)gpa(libopenal.dll,"alSourceQueueBuffers");
	alSourceUnqueueBuffers = (void*)gpa(libopenal.dll,"alSourceUnqueueBuffers");
	alBufferData = (void*)gpa(libopenal.dll,"alBufferData");

	if (er) {
#ifdef _WIN32
		FreeLibrary(libopenal.dll);
#else
		dlclose(libopenal.dll);
#endif
		libopenal.dll = NULL;
		params->call_logmsg("Failed fetching required symbols from " DLLNAME ".");
		return jfauderr_notinited;
	}
#endif

	if (alcIsExtensionPresent(NULL,(ALubyte*)"ALC_ENUMERATION_EXT") == AL_TRUE) {
		char *defdev, *devlist, *devp;

		params->call_logmsg("OpenAL device enumeration:");
		i=1;

		defdev = (char*)alcGetString(NULL, ALC_DEFAULT_DEVICE_SPECIFIER);
		devlist = (char*)alcGetString(NULL, ALC_DEVICE_SPECIFIER);
		devp = devlist;
		while (devp[0]) {
			params->call_logmsg(jsprintf("  %d. %s%s%s",
				i, devp, !strcmp(devp,defdev) ? " (default)" : "",
				!strcmp(devp,params->wavedevice) ? " (selected)" : ""));
			devp += strlen(devp)+1;
			i++;
		}
	}

	params->call_logmsg("OpenAL Initialisation");

	aldev = alcOpenDevice((params->wavedevice[0] ? params->wavedevice : NULL));
	if (!aldev) {
		params->call_logmsg("OpenAL error: failed opening device");
		return jfauderr_openerror;
	}

	alctx = alcCreateContext(aldev, NULL);
	if (!alctx) {
		params->call_logmsg(jsprintf("OpenAL error: failed creating context (%s)",
					alcGetString(aldev,alcGetError(aldev))));
		jfaudopenal_uninit();
		return jfauderr_openerror;
	}

	alcGetError(aldev);
	alcMakeContextCurrent(alctx);
	if ((err = alcGetError(aldev)) != ALC_NO_ERROR) {
		params->call_logmsg(jsprintf("OpenAL error: failed making context current (%s)",alcGetString(aldev,err)));
		jfaudopenal_uninit();
		return jfauderr_openerror;
	}

	params->call_logmsg(jsprintf("  Vendor: %s", alGetString(AL_VENDOR)));
	params->call_logmsg(jsprintf("  Version: %s", alGetString(AL_VERSION)));
	params->call_logmsg(jsprintf("  Renderer: %s", alGetString(AL_RENDERER)));
	params->call_logmsg(jsprintf("  Extensions: %s", alGetString(AL_EXTENSIONS)));
	
	sources = (source *)malloc(sizeof(source) * params->maxvoices);
	if (!sources) {
		jfaudopenal_uninit();
		return jfauderr_outofmemory;
	}

	memset(sources, 0, sizeof(source) * params->maxvoices);
	
	alGetError();
	for (i=params->maxvoices-1;i>=0;i--) {
		alGenSources(1, &sources[i].source);
		if ((err = alGetError()) != AL_NO_ERROR) {
			params->call_logmsg(jsprintf("OpenAL warning: failed generating source (%s)",alGetString(err)));
			//jfaudopenal_uninit();
			//return jfauderr_outofmemory;
			params->maxvoices -= 1+i;
			break;
		}
	}
	params->call_logmsg(jsprintf("%d voice(s) available",params->maxvoices));

	return jfauderr_ok;
}


jfauderr jfaudopenal_uninit(void)
{
	int i;

#if (DYNAMIC_OPENAL == 1)
	if (libopenal.dll) {
#endif

	if (sources) {
		for (i=params->maxvoices-1;i>=0;i--) {
			jfaudopenal_releasesource(i);
			alDeleteSources(1,&sources[i].source);
		}
		free(sources);
		sources = NULL;
	}

	if (alctx) {
		alcMakeContextCurrent(NULL);
		alcDestroyContext(alctx);
		alctx = NULL;
	}

	if (aldev) {
		alcCloseDevice(aldev);
		aldev = NULL;
	}
	
#if (DYNAMIC_OPENAL == 1)
#ifdef _WIN32
	FreeLibrary(libopenal.dll);
#else
	dlclose(libopenal.dll);
#endif
	libopenal.dll = NULL;
	}
#endif

	return jfauderr_ok;
}

jfauderr jfaudopenal_allocsource(int *handle, char stream)
{
	int h;

	testifloaded();
	
	if (!sources) return jfauderr_notinited;

	for (h=params->maxvoices-1; h>=0; h--) {
		if (!sources[h].used) {
			*handle = h;
			sources[h].used = 1;
			sources[h].stream = stream;
			memset(sources[h].buffers, 0, sizeof(sources[h].buffers));
			memset(sources[h].buffersused, 0, sizeof(sources[h].buffersused));

			debuglog(params->call_logmsg,jsprintf("alout: Source %d allocated, stream=%d",h,stream));
			return jfauderr_ok;
		}
	}

	return jfauderr_genericerror;
}

jfauderr jfaudopenal_releasesource(int handle)
{
	ALenum err;
	int i,j;

	testifloaded();
	
	if (!sources) return jfauderr_notinited;
	if (handle < 0 || handle >= params->maxvoices) return jfauderr_badhandle;
	if (!sources[handle].used) return jfauderr_ok;

	alGetError();
	alSourceStop(sources[handle].source);
	if ((err = alGetError()) != AL_NO_ERROR)
		params->call_logmsg(jsprintf("alSourceStop() error: %s",alGetString(err)));

	debuglog(params->call_logmsg,jsprintf("alout: Source %d released",handle));
	
	// free any stream buffers in use
	if (sources[handle].stream) {
		ALint numbufs;
		ALuint buf[STREAMBUFS];

		alGetSourcei(sources[handle].source, AL_BUFFERS_QUEUED, &numbufs);
		for (;numbufs>0;numbufs-=STREAMBUFS)
			alSourceUnqueueBuffers(sources[handle].source, min(STREAMBUFS,numbufs), buf);

		// free the buffers
		for (i=STREAMBUFS-1;i>=0;i--) {
			if (!sources[handle].buffersused[i]) continue;
			jfaudopenal_releasebuffer(&sources[handle].buffers[i]);
		}
	}
	alSourcei(sources[handle].source, AL_BUFFER, 0);	// clear any buffer attached

	sources[handle].used = 0;
	sources[handle].stream = 0;
	memset(sources[handle].buffers, 0, sizeof(sources[handle].buffers));
	memset(sources[handle].buffersused, 0, sizeof(sources[handle].buffersused));
	
	return jfauderr_ok;
}

jfauderr jfaudopenal_preparebuffer(jfaud_buffer *buf)
{
	ALuint bufid;
	ALenum err;

	testifloaded();

	if (buf->albuf != 0) return jfauderr_ok;	// already prepared

	alGetError();
	alGenBuffers(1, &bufid);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alGenBuffers() error: %s",alGetString(err)));
		return jfauderr_genericerror;
	}

	alBufferData(bufid,
		(buf->channels == 1 ?
			(buf->bitspersample == 8 ? AL_FORMAT_MONO8 : AL_FORMAT_MONO16) :
			(buf->bitspersample == 8 ? AL_FORMAT_STEREO8 : AL_FORMAT_STEREO16)
		), buf->data, buf->numsamples * buf->channels * buf->bitspersample / 8, buf->samplerate);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alBufferData() error: %s",alGetString(err)));
		return jfauderr_genericerror;
	}

	jfaudbuffer_free(buf);
	buf->albuf = (unsigned int)bufid;
	debuglog(params->call_logmsg,jsprintf("alout: Prepared buffer %d",bufid));

	return jfauderr_ok;
}

jfauderr jfaudopenal_releasebuffer(jfaud_buffer *buf)
{
	ALenum err;

	testifloaded();

	if (!buf->albuf) return jfauderr_ok;		// not prepared

	alGetError();
	alDeleteBuffers(1, (ALuint*)&buf->albuf);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alDeleteBuffers() error: %s",alGetString(err)));
		return jfauderr_genericerror;
	}

	debuglog(params->call_logmsg,jsprintf("alout: Released buffer %d",buf->albuf));
	buf->albuf = 0;

	return jfauderr_ok;
}

jfauderr jfaudopenal_queuedata(int handle, jfaud_buffer *buf)
{
	int sbuf;
	ALenum err;
	
	testifloaded();
	
	if (!sources) return jfauderr_notinited;
	if (handle < 0 || handle >= params->maxvoices) return jfauderr_badhandle;
	if (!sources[handle].used) return jfauderr_badhandle;

	if (!buf->albuf) return jfauderr_notprepared;

	if (sources[handle].stream) {
		int bufn;
		sbuf = -1;
		for (bufn=STREAMBUFS-1;bufn>=0;bufn--)
			if (!sources[handle].buffersused[bufn]) { sbuf = bufn; break; }
		if (bufn < 0) return jfauderr_genericerror;	// all buffers filled
	} else {
		// check if the first buffer is empty
		if (sources[handle].buffersused[0]) return jfauderr_genericerror;
		sbuf = 0;
	}
	
	if (buf->channels > 2 || (buf->bitspersample != 8 && buf->bitspersample != 16))
		return jfauderr_unrecognised;

	memcpy(&sources[handle].buffers[sbuf], buf, sizeof(jfaud_buffer));
	sources[handle].buffersused[sbuf] = 1;

	// now add the buffer to the source in the appropriate manner
	alGetError();
	if (sources[handle].stream) {
		alSourceQueueBuffers(sources[handle].source, 1, &buf->albuf);
		if ((err = alGetError()) != AL_NO_ERROR) {
			params->call_logmsg(jsprintf("alSourceQueueBuffers() error: %s",alGetString(err)));
			return jfauderr_genericerror;
		}

		if (sources[handle].supposedly == jfaud_playing) {
			// this happens is we hitch for some reason and run out of data during, say,
			// a mode change or other halt-inducing event
			ALint val = jfaud_stopped;
			
			alGetError();
			alGetSourcei(sources[handle].source, AL_SOURCE_STATE, &val);
			alGetError();

			if (val != AL_PLAYING) alSourcePlay(sources[handle].source);
		}
				
		debuglog(params->call_logmsg,jsprintf("alout: Queued buffer on source %d as buf[%d]",handle,sbuf));
	} else {
		alSourcei(sources[handle].source, AL_BUFFER, buf->albuf);
		if ((err = alGetError()) != AL_NO_ERROR) {
			params->call_logmsg(jsprintf("alSourcei(AL_BUFFER) error: %s",alGetString(err)));
			return jfauderr_genericerror;
		}

		alSourceRewind(sources[handle].source);
		if ((err = alGetError()) != AL_NO_ERROR) {
			params->call_logmsg(jsprintf("alSourceRewind() error: %s",alGetString(err)));
			return jfauderr_genericerror;
		}

		debuglog(params->call_logmsg,jsprintf("alout: Attached buffer to source %d",handle));
	}
	
	return jfauderr_ok;
}

jfauderr jfaudopenal_needsdata(int handle)
{
	int i;

	testifloaded();

	if (!sources) return jfauderr_notinited;
	if (handle < 0 || handle >= params->maxvoices) return jfauderr_badhandle;
	if (!sources[handle].used) return jfauderr_badhandle;

	if (sources[handle].stream) {
		// check if there's any empty buffers available for filling
		for (i=STREAMBUFS-1;i>=0;i--)
			if (!sources[handle].buffersused[i]) return jfauderr_ok;
	} else {
		// check if the first buffer is empty
		if (!sources[handle].buffersused[0]) return jfauderr_ok;
	}

	return jfauderr_genericerror;
}

jfauderr jfaudopenal_update(void)
{
	ALenum err;
	int i,j,k;
	ALint numbufs;
	ALuint buf;
	
	// does housekeeping
	testifloaded();

	alGetError();
	
	// dequeues played buffers in played streams, etc
	for (i=params->maxvoices-1; i>=0; i--) {
		if (!sources[i].used) continue;
		if (sources[i].stream) {
			alGetSourcei(sources[i].source, AL_BUFFERS_PROCESSED, &numbufs);
			if ((err = alGetError()) != AL_NO_ERROR) {
				params->call_logmsg(jsprintf("alGetSourcei(AL_BUFFERS_PROCESSED) error: %s",alGetString(err)));
				continue;
			}

			for (;numbufs>0;numbufs--) {
				alSourceUnqueueBuffers(sources[i].source, 1, &buf);
				if ((err = alGetError()) != AL_NO_ERROR) {
					params->call_logmsg(jsprintf("alSourceUnqueueBuffers() error: %s",alGetString(err)));
					break;
				}
				for (j=STREAMBUFS-1;j>=0;j--) {
					if (!sources[i].buffersused[j]) continue;
					if (sources[i].buffers[j].albuf != buf) continue;
					
					jfaudopenal_releasebuffer(&sources[i].buffers[j]);	// frees the buffer
					sources[i].buffersused[j] = 0;
					debuglog(params->call_logmsg,jsprintf("alout: Removed buffer from source %d",i));
					break;
				}

				// buffer was queued on source, but we have no record of it!?
				if (j<0) debuglog(params->call_logmsg,jsprintf("alout: Dequeued buffer from source %d, but not knowingly attached to begin with!",i));
			}
		}
	}

	return jfauderr_ok;
}

jfauderr jfaudopenal_setsourceprops(int handle, float volume, float pitch, char loop, char threed)
{
	ALenum err;
	jfauderr rv;

	testifloaded();
	
	if (!sources) return jfauderr_notinited;
	if (handle < 0 || handle >= params->maxvoices) return jfauderr_badhandle;
	if (!sources[handle].used) return jfauderr_badhandle;

	alGetError();
	rv = jfauderr_ok;

	alSourcef(sources[handle].source, AL_GAIN, volume);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alSourcef(AL_GAIN) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	alSourcef(sources[handle].source, AL_PITCH, pitch);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alSourcef(AL_PITCH) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	alSourcei(sources[handle].source, AL_LOOPING, loop ? AL_TRUE : AL_FALSE);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alSourcei(AL_LOOPING) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	// 2D voices don't diminish with distance unless their gain is adjusted
	alSourcef(sources[handle].source, AL_ROLLOFF_FACTOR, threed ? 1.0 : 0.0);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alSourcei(AL_ROLLOFF_FACTOR) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	return rv;
}

jfauderr jfaudopenal_setsourcepos(int handle,
		float posx, float posy, float posz,
		float velx, float vely, float velz,
		char posrel)
{
	ALenum err;
	jfauderr rv;

	testifloaded();
	
	if (!sources) return jfauderr_notinited;
	if (handle < 0 || handle >= params->maxvoices) return jfauderr_badhandle;
	if (!sources[handle].used) return jfauderr_badhandle;

	alGetError();
	rv = jfauderr_ok;

	alSourcei(sources[handle].source, AL_SOURCE_RELATIVE, posrel ? AL_TRUE : AL_FALSE);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alSourcei(AL_SOURCE_RELATIVE) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	alSource3f(sources[handle].source, AL_POSITION, posx, posy, posz);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alSource3f(AL_POSITION) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	return rv;
}

jfauderr jfaudopenal_getsourceplaymode(int handle, jfaudplaymode *mode)
{
	ALenum err;
	ALint val;

	testifloaded();
	
	if (!sources) return jfauderr_notinited;
	if (handle < 0 || handle >= params->maxvoices) return jfauderr_badhandle;
	if (!sources[handle].used) return jfauderr_badhandle;

	alGetError();
	alGetSourcei(sources[handle].source, AL_SOURCE_STATE, &val);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alGetSourcei(AL_SOURCE_STATE) error: %s",alGetString(err)));
		return jfauderr_genericerror;
	}
	
	switch (val) {
		case AL_PLAYING: *mode = jfaud_playing; break;
		case AL_PAUSED: *mode = jfaud_paused; break;
		default: *mode = jfaud_stopped; break;
	}

	sources[handle].supposedly = *mode;

	return jfauderr_ok;
}

jfauderr jfaudopenal_setsourceplaymode(int handle, jfaudplaymode mode)
{
	ALenum err;
	char *act;

	testifloaded();
	
	if (!sources) return jfauderr_notinited;
	if (handle < 0 || handle >= params->maxvoices) return jfauderr_badhandle;
	if (!sources[handle].used) return jfauderr_badhandle;

	alGetError();
	
	switch (mode) {
		case jfaud_playing: alSourcePlay(sources[handle].source); act="Play"; break;
		case jfaud_paused: alSourcePause(sources[handle].source); act="Pause"; break;
		default: alSourceStop(sources[handle].source); act="Stop"; break;
	}

	sources[handle].supposedly = mode;

	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alSource%s() error: %s",act,alGetString(err)));
		return jfauderr_genericerror;
	}
	
	return jfauderr_ok;
}

jfauderr jfaudopenal_setpos(float posx, float posy, float posz,
		float velx, float vely, float velz,
		float atx, float aty, float atz,
		float upx, float upy, float upz)
{
	ALenum err;
	jfauderr rv;
	float orientation[6];
	
	testifloaded();

	alGetError();
	rv = jfauderr_ok;

	alListener3f(AL_POSITION, posx, posy, posz);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alListener3f(AL_POSITION) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	alListener3f(AL_VELOCITY, velx, vely, velz);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alListener3f(AL_VELOCITY) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	orientation[0] = atx;
	orientation[1] = aty;
	orientation[2] = atz;
	orientation[3] = upx;
	orientation[4] = upy;
	orientation[5] = upz;
	alListenerfv(AL_ORIENTATION, orientation);
	if ((err = alGetError()) != AL_NO_ERROR) {
		params->call_logmsg(jsprintf("alListenerfv(AL_ORIENTATION) error: %s",alGetString(err)));
		rv = jfauderr_genericerror;
	}

	return rv;
}

