#include "internal.h"

#include <stdlib.h>
#include <stdio.h>

static struct jfaud_init_params *params = NULL;

static jfauderr init(void)
{
	params = jfaud_getparams();
	return jfauderr_ok;
}

static jfauderr uninit(void)
{
	jfaudmidi_stop();
	return jfauderr_ok;
}

static jfauderr testfile(jfaud_file *file, const char *sub)
{
	midiseq *seq;
	
	seq = new midiseq;
	if (!seq) return jfauderr_outofmemory;
	
	params->call_seek(file->fh, 0, SEEK_SET);
	
	switch (seq->loadfile(file->fh)) {
		case 0:
			if (seq->setupsequencer(midicardtype, 0)) {
				delete seq;
				return jfauderr_genericerror;
			}

			params->call_close(file->fh);	// we don't need the file handle
			
			file->type = jfaudfile_midi;
			file->readerinfo = (void*)seq;

			return jfauderr_ok;
		case -1:
			delete seq;
			return jfauderr_genericerror;
		case -2:
			delete seq;
			return jfauderr_unrecognised;
		default: break;
	}
	return jfauderr_unrecognised;
}

static jfauderr cleanup(jfaud_file *file, char *noclose)
{
	jfaudmidi_stop();

	if (file->readerinfo) {
		delete (midiseq *)file->readerinfo;
	}
	file->readerinfo = NULL;
	*noclose = 1;	// we closed the handle earlier

	return jfauderr_ok;
}

static jfauderr getplaytime(jfaud_file *file, double *len)
{
	*len = 3600.0;	// FIXME: work out the length
	return jfauderr_ok;
}

extern "C" {
jfaud_filereader reader_midi = {
	"MIDI file",
	jfaudfile_midi,
	jfaudfiledata_midi,
	
	init,
	uninit,
	testfile,
	cleanup,
	
	NULL,NULL,NULL,NULL, getplaytime
};
}
