#include "internal.h"


static unsigned long crc32table[256];
static char inited = 0;

static void initcrc32table(void)
{
	unsigned long i,j,k;
	
	// algorithm and polynomial same as that used by infozip's zip
	for (i=0; i<256; i++) {
		j = i;
		for (k=8; k; k--)
			j = (j&1) ? (0xedb88320L ^ (j>>1)) : (j>>1);
		crc32table[i] = j;
	}

	inited=1;
}


static void crc32init(unsigned long *crcvar)
{
	if (!crcvar) return;
	*crcvar = 0xffffffffl;
}

static void crc32block(unsigned long *crcvar, const unsigned char *blk, unsigned long len)
{
	unsigned long crc = *crcvar;
	while (len--) crc = crc32table[(crc ^ *(blk++)) & 0xffl] ^ (crc >> 8);
	*crcvar = crc;
}

static unsigned long crc32finish(unsigned long *crcvar)
{
	*crcvar = *crcvar ^ 0xffffffffl;
	return *crcvar;
}

jfaudident jfaud_hashfn(const unsigned char *blk)
{
	unsigned long crc, len;

	if (!inited) initcrc32table();
	
	/*
	crc32init(&crc);
	crc32block(&crc, blk, strlen(blk));
	return crc32finish(&crc);
	*/
	crc = 0xffffffffl;
	len = strlen(blk);
	while (len--) crc = crc32table[(crc ^ *(blk++)) & 0xffl] ^ (crc >> 8);
	crc ^= 0xffffffffl;
	return (jfaudident)crc;
}


