#include <string.h>
#include "internal.h"


#define FIXSHIFT 8
jfauderr fx_nearest(PcmBuffer *buf)
{
	jfauderr err;
	unsigned long rateratio, destlen, i, j;

	PcmBuffer dst;

	if (buf->bitspersample != 8 && buf->bitspersample != 16) return jfauderr_notcompatible;
	if (buf->samplerate >= conf.samplerate) return jfauderr_ok;	// we don't downsample

	rateratio = (buf->samplerate << FIXSHIFT) / conf.samplerate;
	destlen = (buf->numsamples * ((conf.samplerate << FIXSHIFT) / buf->samplerate)) >> FIXSHIFT;

	if ((err = pcmbuffer_allocate(&dst, destlen, conf.samplerate, buf->channels, buf->bitspersample)) != jfauderr_ok)
		return jfauderr_outofmemory;

	destlen *= buf->channels;

	if (buf->bitspersample == 8) {
		unsigned char *bptrs, *bptrd;
		bptrs = (unsigned char *)buf->data;
		bptrd = (unsigned char *)dst.data;
		for (i=j=0; j < destlen; i+=rateratio,j++)
			bptrd[j] = bptrs[i>>FIXSHIFT];
	} else {
		unsigned short *wptrs, *wptrd;
		wptrs = (unsigned short *)buf->data;
		wptrd = (unsigned short *)dst.data;
		for (i=j=0; j < destlen; i+=rateratio,j++)
			wptrd[j] = wptrs[i>>FIXSHIFT];
	}

	pcmbuffer_free(buf);
	memcpy(buf, &dst, sizeof(dst));

	return jfauderr_ok;
}

jfauderr fx_stertomono(PcmBuffer *buf)
{
	int s;
	unsigned int n;

	if (buf->channels == 1) return jfauderr_noerror;

	if (buf->bitspersample == 8) {
		unsigned char *src = (unsigned char *)buf->data;
		unsigned char *dst = (unsigned char *)buf->data;

		for (n = buf->numsamples; n>0; n--) {
			s = src[0] + src[1];
			s >>= 1;

			if (s < 0) s = 0;
			else if (s > 255) s = 255;

			*dst = (unsigned char)s;

			src += 2;
			dst += 1;
		}
	} else {
		signed short *src = (signed short *)buf->data;
		signed short *dst = (signed short *)buf->data;

		for (n = buf->numsamples; n>0; n--) {
			s = src[0] + src[1];
			s >>= 1;

			if (s < -32768) s = -32768;
			else if (s > 32767) s = 32767;

			*dst = (signed short)s;

			src += 2;
			dst += 1;
		}
	}

	buf->channels = 1;

	return jfauderr_noerror;
}

