#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "internal.h"

#if (WITH_FLAC == 0)
static jfauderr init(void) { return jfauderr_disabled; }
static jfauderr uninit(void) { return jfauderr_disabled; }
static jfauderr test(File *file UNUSED) { return jfauderr_disabled; }
static jfauderr unload(File *file UNUSED, char *noclose UNUSED) { return jfauderr_disabled; }
static jfauderr getmetainfo(File *file UNUSED, char what UNUSED, char *buf UNUSED, int len UNUSED) { return jfauderr_disabled; }
static jfauderr readsamples(File *file UNUSED, int n UNUSED, PcmBuffer *buf UNUSED, char loop UNUSED) { return jfauderr_disabled; }
static jfauderr readallsamples(File *file UNUSED, PcmBuffer *buf UNUSED) { return jfauderr_disabled; }
static jfauderr getplaytime(File *file UNUSED, float *len UNUSED) { return jfauderr_disabled; }
#else

#include "seekable_stream_decoder.h"

#if (DYNAMIC_FLAC == 1)

#ifdef _WIN32
#ifdef PACKED
#undef PACKED
#endif
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define DLLNAME "libflac.dll"
#else

#include <dlfcn.h>
#define DLLNAME "libFLAC.so"
#endif


static struct {
#ifdef _WIN32
        HMODULE dll;
#else

        void *dll;
#endif

	FLAC__SeekableStreamDecoder *(*FLAC__seekable_stream_decoder_new)();
	void (*FLAC__seekable_stream_decoder_delete)(FLAC__SeekableStreamDecoder *decoder);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_client_data)(FLAC__SeekableStreamDecoder *decoder, void *value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_read_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderReadCallback value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_write_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderWriteCallback value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_seek_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderSeekCallback value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_tell_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderTellCallback value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_length_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderLengthCallback value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_eof_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderEofCallback value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_metadata_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderMetadataCallback value);
	FLAC__bool (*FLAC__seekable_stream_decoder_set_error_callback)(FLAC__SeekableStreamDecoder *decoder, FLAC__SeekableStreamDecoderErrorCallback value);
	FLAC__SeekableStreamDecoderState (*FLAC__seekable_stream_decoder_init)(FLAC__SeekableStreamDecoder *decoder);
	FLAC__bool (*FLAC__seekable_stream_decoder_finish)(FLAC__SeekableStreamDecoder *decoder);
	FLAC__SeekableStreamDecoderState (*FLAC__seekable_stream_decoder_get_state)(const FLAC__SeekableStreamDecoder *decoder);
	const char *(*FLAC__seekable_stream_decoder_get_resolved_state_string)(const FLAC__SeekableStreamDecoder *decoder);
	FLAC__bool (*FLAC__seekable_stream_decoder_process_single)(FLAC__SeekableStreamDecoder *decoder);
	FLAC__bool (*FLAC__seekable_stream_decoder_process_until_end_of_metadata)(FLAC__SeekableStreamDecoder *decoder);
	FLAC__bool (*FLAC__seekable_stream_decoder_seek_absolute)(FLAC__SeekableStreamDecoder *decoder, FLAC__uint64 sample);

	const char * const *FLAC__SeekableStreamDecoderStateString;
} libflac;

#define FLAC__seekable_stream_decoder_new libflac.FLAC__seekable_stream_decoder_new
#define FLAC__seekable_stream_decoder_delete libflac.FLAC__seekable_stream_decoder_delete
#define FLAC__seekable_stream_decoder_set_client_data libflac.FLAC__seekable_stream_decoder_set_client_data
#define FLAC__seekable_stream_decoder_set_read_callback libflac.FLAC__seekable_stream_decoder_set_read_callback
#define FLAC__seekable_stream_decoder_set_write_callback libflac.FLAC__seekable_stream_decoder_set_write_callback
#define FLAC__seekable_stream_decoder_set_seek_callback libflac.FLAC__seekable_stream_decoder_set_seek_callback
#define FLAC__seekable_stream_decoder_set_tell_callback libflac.FLAC__seekable_stream_decoder_set_tell_callback
#define FLAC__seekable_stream_decoder_set_length_callback libflac.FLAC__seekable_stream_decoder_set_length_callback
#define FLAC__seekable_stream_decoder_set_eof_callback libflac.FLAC__seekable_stream_decoder_set_eof_callback
#define FLAC__seekable_stream_decoder_set_metadata_callback libflac.FLAC__seekable_stream_decoder_set_metadata_callback
#define FLAC__seekable_stream_decoder_set_error_callback libflac.FLAC__seekable_stream_decoder_set_error_callback
#define FLAC__seekable_stream_decoder_init libflac.FLAC__seekable_stream_decoder_init
#define FLAC__seekable_stream_decoder_finish libflac.FLAC__seekable_stream_decoder_finish
#define FLAC__seekable_stream_decoder_get_state libflac.FLAC__seekable_stream_decoder_get_state
#define FLAC__seekable_stream_decoder_get_resolved_state_string libflac.FLAC__seekable_stream_decoder_get_resolved_state_string
#define FLAC__seekable_stream_decoder_process_single libflac.FLAC__seekable_stream_decoder_process_single
#define FLAC__seekable_stream_decoder_process_until_end_of_metadata libflac.FLAC__seekable_stream_decoder_process_until_end_of_metadata
#define FLAC__seekable_stream_decoder_seek_absolute libflac.FLAC__seekable_stream_decoder_seek_absolute
#define FLAC__SeekableStreamDecoderStateString libflac.FLAC__SeekableStreamDecoderStateString

#define testifloaded() { if (libflac.dll == NULL) return jfauderr_notinited; }
#else

#define testifloaded() { }
#endif

typedef struct {
	WaveformFile_defn

	unsigned samplerate, channels;
	FLAC__uint64 filelen, pcmlen;

	short *pcm;
	unsigned pcmleft, blocksize;
	
	FLAC__SeekableStreamDecoder *ssd;
} Flac;

static FLAC__SeekableStreamDecoderReadStatus flacReadCallback(
		const FLAC__SeekableStreamDecoder *decoder UNUSED,
		FLAC__byte buffer[], unsigned *bytes, void *client_data)
{
	File *file = (File *)client_data;
	int rb;

	rb = sysf_read(&file->fh, buffer, *bytes);
	if (rb < 0) return FLAC__SEEKABLE_STREAM_DECODER_READ_STATUS_ERROR;

	*bytes = (unsigned)rb;

	return FLAC__SEEKABLE_STREAM_DECODER_READ_STATUS_OK;
}

static FLAC__StreamDecoderWriteStatus flacWriteCallback(
		const FLAC__SeekableStreamDecoder *decoder UNUSED, const FLAC__Frame *frame,
		const FLAC__int32 *const buffer[], void *client_data)
{
	Flac *flac = (Flac *)(((File *)client_data)->sf);
	unsigned i,j;
	short *p;

	/*
	flac->samplerate = frame->header.sample_rate;
	flac->channels = frame->header.channels;
	flac->blocksize = (int)frame->header.channels * 2;
	*/
	flac->pcmleft = (int)frame->header.blocksize;
	if (flac->pcm) free(flac->pcm);
	flac->pcm = (short *)malloc(flac->pcmleft * flac->blocksize);
	if (!flac->pcm) { flac->pcmleft = 0; return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT; }

	p = flac->pcm;
	for (i = 0; i < frame->header.blocksize; i++)
		for (j = 0; j < frame->header.channels; j++)
			*(p++) = buffer[j][i];// >> 16;

	return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

static FLAC__SeekableStreamDecoderSeekStatus flacSeekCallback(
		const FLAC__SeekableStreamDecoder *decoder UNUSED,
		FLAC__uint64 absolute_byte_offset, void *client_data)
{
	File *file = (File *)client_data;

	if (sysf_seek(&file->fh, (int)absolute_byte_offset, SEEK_SET) < 0)
		return FLAC__SEEKABLE_STREAM_DECODER_SEEK_STATUS_ERROR;

	return FLAC__SEEKABLE_STREAM_DECODER_SEEK_STATUS_OK;
}

static FLAC__SeekableStreamDecoderTellStatus flacTellCallback(
		const FLAC__SeekableStreamDecoder *decoder UNUSED,
		FLAC__uint64 *absolute_byte_offset, void *client_data)
{
	File *file = (File *)client_data;
	int rb;

	rb = sysf_tell(&file->fh);
	if (rb < 0) return FLAC__SEEKABLE_STREAM_DECODER_TELL_STATUS_ERROR;

	*absolute_byte_offset = (FLAC__uint64)rb;
	return FLAC__SEEKABLE_STREAM_DECODER_TELL_STATUS_OK;
}

static FLAC__SeekableStreamDecoderLengthStatus flacLengthCallback(
		const FLAC__SeekableStreamDecoder *decoder UNUSED,
		FLAC__uint64 *stream_length, void *client_data)
{
	Flac *flac = (Flac *)(((File *)client_data)->sf);

	*stream_length = flac->filelen;
	return FLAC__SEEKABLE_STREAM_DECODER_LENGTH_STATUS_OK;
}

static FLAC__bool flacEofCallback(const FLAC__SeekableStreamDecoder *decoder UNUSED, void *client_data)
{
	File *file = (File *)client_data;
	Flac *flac = (Flac *)file->sf;
	int rb;

	rb = sysf_tell(&file->fh);
	if (rb < 0) return 1;

	return ((FLAC__uint64)rb >= flac->filelen);
}

static void flacMetadataCallback(const FLAC__SeekableStreamDecoder *decoder UNUSED,
		const FLAC__StreamMetadata *metadata, void *client_data)
{
	File *file = (File *)client_data;
	Flac *flac = (Flac *)file->sf;
	switch (metadata->type) {
		case FLAC__METADATA_TYPE_STREAMINFO:
			flac->samplerate = metadata->data.stream_info.sample_rate;
			flac->channels   = metadata->data.stream_info.channels;
			flac->pcmlen     = metadata->data.stream_info.total_samples;;
			flac->blocksize  = flac->channels * 2;
			break;
		default: break;
	}
}

static void flacErrorCallback(const FLAC__SeekableStreamDecoder *decoder UNUSED,
		FLAC__StreamDecoderErrorStatus status UNUSED, void *client_data UNUSED)
{
	//File *file = (File *)client_data;
	debuglog(DEBUGLVL_GENERAL,"flacErrorCallback called");
}


// SoundFileReader functions

static jfauderr init(void)
{
#if (DYNAMIC_FLAC == 1)
        int er=0;

#ifdef _WIN32
        void *sym;
        libflac.dll = LoadLibrary(DLLNAME);

#define gpa(d,x) sym = GetProcAddress(d,x); { \
                if (!sym) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#else
        libflac.dll = dlopen(DLLNAME, RTLD_GLOBAL|RTLD_NOW);
        dlerror();

#define gpa(d,x) dlsym(d,x); { \
                const char *s; \
                if ((s = dlerror())) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#endif

        if (!libflac.dll) return jfauderr_disabled;
        
        FLAC__seekable_stream_decoder_new = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_new");
        FLAC__seekable_stream_decoder_delete = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_delete");
        FLAC__seekable_stream_decoder_set_client_data = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_client_data");
        FLAC__seekable_stream_decoder_set_read_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_read_callback");
        FLAC__seekable_stream_decoder_set_write_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_write_callback");
        FLAC__seekable_stream_decoder_set_seek_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_seek_callback");
        FLAC__seekable_stream_decoder_set_tell_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_tell_callback");
        FLAC__seekable_stream_decoder_set_length_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_length_callback");
        FLAC__seekable_stream_decoder_set_eof_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_eof_callback");
        FLAC__seekable_stream_decoder_set_metadata_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_metadata_callback");
	FLAC__seekable_stream_decoder_set_error_callback = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_set_error_callback");
        FLAC__seekable_stream_decoder_init = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_init");
        FLAC__seekable_stream_decoder_finish = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_finish");
        FLAC__seekable_stream_decoder_get_state = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_get_state");
        FLAC__seekable_stream_decoder_get_resolved_state_string = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_get_resolved_state_string");
        FLAC__seekable_stream_decoder_process_single = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_process_single");
        FLAC__seekable_stream_decoder_process_until_end_of_metadata = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_process_until_end_of_metadata");
	FLAC__seekable_stream_decoder_seek_absolute = (void*)gpa(libflac.dll, "FLAC__seekable_stream_decoder_seek_absolute");
	FLAC__SeekableStreamDecoderStateString = (void*)gpa(libflac.dll, "FLAC__SeekableStreamDecoderStateString");
	
        if (er) {
#ifdef _WIN32
                FreeLibrary(libflac.dll);
#else
                dlclose(libflac.dll);
#endif

                libflac.dll = NULL;
                debuglog(DEBUGLVL_GENERAL,"Failed fetching required symbols from " DLLNAME ".");
                return jfauderr_initfailed;
        }
#endif	

	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
#if (DYNAMIC_FLAC == 1)
        if (libflac.dll) {
#ifdef _WIN32
                FreeLibrary(libflac.dll);
#else
                dlclose(libflac.dll);
#endif

                libflac.dll = NULL;
        }
#endif

	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	FLAC__SeekableStreamDecoder *ssd;
	FLAC__SeekableStreamDecoderState ssdstate;
	Flac *flac;
	char sig[4];
	
	flac = (Flac *)calloc(1, sizeof(Flac));
	if (!flac) {
		FLAC__seekable_stream_decoder_delete(ssd);
		return jfauderr_outofmemory;
	}

	file->sf = (SoundFile *)flac;

	sysf_seek(&file->fh, 0, SEEK_END);	// get file length for eof tests
	flac->filelen = sysf_tell(&file->fh);
	sysf_seek(&file->fh, 0, SEEK_SET);

	// I would have expected the FLAC library to be able to do this for me. It doesn't appear to want to.
	if (sysf_read(&file->fh, sig, 4) != 4) {
		free(flac); file->sf = NULL;
		return jfauderr_notcompatible;
	}
	if (sig[0] != 'f' || sig[1] != 'L' || sig[2] != 'a' || sig[3] != 'C') {
		free(flac); file->sf = NULL;
		return jfauderr_notrecognised;
	}
	sysf_seek(&file->fh, 0, SEEK_SET);

	ssd = flac->ssd = FLAC__seekable_stream_decoder_new();
	if (!ssd) {
		free(flac); file->sf = NULL;
		return jfauderr_outofmemory;
	}

	if (!FLAC__seekable_stream_decoder_set_client_data(ssd, (void*)file) ||
	    !FLAC__seekable_stream_decoder_set_read_callback(ssd, flacReadCallback) ||
	    !FLAC__seekable_stream_decoder_set_write_callback(ssd, flacWriteCallback) ||
	    !FLAC__seekable_stream_decoder_set_seek_callback(ssd, flacSeekCallback) ||
	    !FLAC__seekable_stream_decoder_set_tell_callback(ssd, flacTellCallback) ||
	    !FLAC__seekable_stream_decoder_set_length_callback(ssd, flacLengthCallback) ||
	    !FLAC__seekable_stream_decoder_set_eof_callback(ssd, flacEofCallback) ||
	    !FLAC__seekable_stream_decoder_set_metadata_callback(ssd, flacMetadataCallback) ||
	    !FLAC__seekable_stream_decoder_set_error_callback(ssd, flacErrorCallback)) {
		FLAC__seekable_stream_decoder_delete(ssd);
		free(flac); file->sf = NULL;
		return jfauderr_genericerror;
	}

	ssdstate = FLAC__seekable_stream_decoder_init(ssd);
	if (ssdstate != FLAC__SEEKABLE_STREAM_DECODER_OK) {
		debuglog(DEBUGLVL_WORDY,jsprintf("flac error = %s\n", FLAC__SeekableStreamDecoderStateString[ssdstate]));
		FLAC__seekable_stream_decoder_delete(ssd);
		free(flac); file->sf = NULL;
		return jfauderr_genericerror;
	}

	if (!FLAC__seekable_stream_decoder_process_until_end_of_metadata(ssd)) {
		debuglog(DEBUGLVL_WORDY,jsprintf("flac error = %s",FLAC__seekable_stream_decoder_get_resolved_state_string(ssd)));
		FLAC__seekable_stream_decoder_finish(ssd);
		FLAC__seekable_stream_decoder_delete(ssd);
		free(flac); file->sf = NULL;
		return jfauderr_genericerror;
	}
	
	flac->type = type_wave;
	flac->format = fmt_flac;

	return jfauderr_ok;
}

static jfauderr unload(File *file, char *noclose)
{
	Flac *flac = (Flac *)file->sf;

	FLAC__seekable_stream_decoder_finish(flac->ssd);
	FLAC__seekable_stream_decoder_delete(flac->ssd);
	if (flac->pcm) free(flac->pcm);
	
	// caller cleans up file->sf
	*noclose = 0;

	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file UNUSED, JFAudMetaInfo what UNUSED, char *buf UNUSED, int *len UNUSED)
{
	return jfauderr_notimpl;
}

// WaveformFileReader functions
static jfauderr loadsound(File *file UNUSED, Flac *flac, PcmBuffer *buf, unsigned long samples, char loop)
{
	jfauderr err;
	char *rp;
	unsigned int samps;

	err = pcmbuffer_allocate(buf, samples, flac->samplerate, flac->channels, 16);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;
	
	rp = (char *)buf->data;
	do {
		if (flac->pcmleft == 0) {
			if (FLAC__seekable_stream_decoder_get_state(flac->ssd) ==
					FLAC__SEEKABLE_STREAM_DECODER_END_OF_STREAM) {
				if (loop) {
					if (!FLAC__seekable_stream_decoder_seek_absolute(flac->ssd, 0))
						return jfauderr_ioerror;
					continue;
				} else {
					break;
				}
			} else if (!FLAC__seekable_stream_decoder_process_single(flac->ssd)) {
				return jfauderr_genericerror;
			}
		} else {
			samps = min(samples, flac->pcmleft);

			memcpy(rp, flac->pcm, samps * flac->blocksize);
			flac->pcmleft -= samps;
			memmove(flac->pcm, (char*)flac->pcm + samps * flac->blocksize,
					flac->pcmleft * flac->blocksize);

			rp += samps * flac->blocksize;
			samples -= samps;
			buf->numsamples += samps;
		}
	} while (samples > 0);

	if (buf->numsamples == 0)
		pcmbuffer_free(buf);

	return jfauderr_noerror;
}

static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	Flac *flac;

	flac = (Flac *)file->sf;

	if (n <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	return loadsound(file, flac, buf, n, loop);
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	Flac *flac = (Flac *)file->sf;

	if (!FLAC__seekable_stream_decoder_seek_absolute(flac->ssd, 0)) return jfauderr_ioerror;

	return loadsound(file, flac, buf, flac->pcmlen, 0);
}

static jfauderr getplaytime(File *file, float *len)
{
	Flac *flac;

	flac = (Flac *)file->sf;

	if (flac->pcmlen == 0) *len = -1.0;	// unspecified file length in header
	else *len = (float)flac->pcmlen / (float)flac->samplerate;
	
	return jfauderr_noerror;
}

#endif

WaveformFileReader wf_flac = {
	"FLAC",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_flac,
	readsamples,
	readallsamples,
	getplaytime
};

