#if (DYNAMIC_MOD == 1)
#error ERROR ERROR ERROR: Since Modplug is C++ and I haven''t made my own C wrapper around it, we link to it at compile time.
#endif

#include "stdafx.h"
#include "sndfile.h"

extern "C" {

#include "internal.h"

typedef struct {
	WaveformFile_defn

	CSoundFile *mod;
} ModPlug;

// The largest module in my collection is aws_fan2.xm at 5368865 bytes.
// This is so people don't get clever and try giving us a 200MB
// file or something, knowing the last resort test (currently this MOD
// loader) will require it be loaded entirely into memory
#define LARGESTMOD (1048576*6)


// SoundFileReader functions

static jfauderr init(void)
{
	CSoundFile::SetReverbParameters(0,0);	// depth, delay
	CSoundFile::SetXBassParameters(0,0);	// amount, range
	CSoundFile::SetSurroundParameters(0,0);	// depth, delay

	CSoundFile::SetWaveConfig(min(conf.samplerate,44100), 16, 2, 1);	// samplerate, bits, channels, mmx

	// surround, no-oversampling, reverb, hq-resampler, megabass, noise-reduction, eq
	CSoundFile::SetWaveConfigEx(0, 0, 0, 1, 0, 0, 0);
	CSoundFile::SetResamplingMode(1);	// 0=nearest, 1=linear, 2=spline, 3=fir

	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	ModPlug mod;
	void *data;
	int length;

	if (!file) return jfauderr_genericerror;
	if (file->sf) return jfauderr_genericerror;	// already been identified

	mod.type = type_wave;
	mod.format = fmt_mod;

	sysf_seek(&file->fh, 0, SEEK_SET);

	length = sysf_filelen(&file->fh);
	if (length < 0) return jfauderr_notrecognised;
	if (length > LARGESTMOD) return jfauderr_notrecognised;

	data = malloc(length);
	if (!data) return jfauderr_outofmemory;

	if (sysf_read(&file->fh, data, length) != length) {
		free(data);
		return jfauderr_notrecognised;
	}

	mod.mod = new CSoundFile;
	if (!mod.mod) {
		free(data);
		return jfauderr_outofmemory;	// FIXME: this should really check if out of memory or something...
	}

	if (mod.mod->Create((const BYTE*)data, length)) {
		int reject = 0;
		const char *modtype = "Unknown";;
		switch (mod.mod->GetType()) {
			case MOD_TYPE_NONE: modtype = "None"; break;
			case MOD_TYPE_MOD: if (file->filename) {
				// damn near anything can be identified as a valid MOD,
				// so do an extension check
				char *dot;
				dot = strrchr(file->filename, '.');
				if (!dot) reject = 1;
				else if (strcasecmp(dot,".mod")) reject = 1;
			} else reject = 1;
				modtype = "MOD";
				break;
			case MOD_TYPE_S3M:  modtype = "S3M"; break;
			case MOD_TYPE_XM:   modtype = "XM"; break;
			case MOD_TYPE_MED:  modtype = "MED"; break;
			case MOD_TYPE_MTM:  modtype = "MTM"; break;
			case MOD_TYPE_IT:   modtype = "IT"; break;
			case MOD_TYPE_669:  modtype = "669"; break;
			case MOD_TYPE_ULT:  modtype = "ULT"; break;
			case MOD_TYPE_STM:  modtype = "STM"; break;
			case MOD_TYPE_FAR:  modtype = "FAR"; break;
			case MOD_TYPE_WAV:  modtype = "WAV"; reject = 1; break;
			case MOD_TYPE_AMF:  modtype = "AMF"; break;
			case MOD_TYPE_AMS:  modtype = "AMS"; break;
			case MOD_TYPE_DSM:  modtype = "DSM"; break;
			case MOD_TYPE_MDL:  modtype = "MDL"; break;
			case MOD_TYPE_OKT:  modtype = "OKT"; break;
			case MOD_TYPE_MID:  modtype = "MID"; reject = 1; break;
			case MOD_TYPE_DMF:  modtype = "DMF"; break;
			case MOD_TYPE_PTM:  modtype = "PTM"; break;
			case MOD_TYPE_DBM:  modtype = "DBM"; break;
			case MOD_TYPE_MT2:  modtype = "MT2"; break;
			case MOD_TYPE_AMF0: modtype = "AMF0"; break;
			case MOD_TYPE_PSM:  modtype = "PSM"; break;
			case MOD_TYPE_J2B:  modtype = "J2B"; break;
			case MOD_TYPE_UMX:  modtype = "UMX"; break;
			default: modtype = "Unknown"; break;
		}
		if (reject) {
			debuglog(DEBUGLVL_WORDY,jsprintf("type = %s, but rejected",modtype));
			delete mod.mod;
			free(data);
			return jfauderr_notrecognised;
		} else debuglog(DEBUGLVL_WORDY,jsprintf("type = %s",modtype));
	} else {
		delete mod.mod;
		free(data);
		return jfauderr_notrecognised;
	}

	file->sf = (SoundFile *)malloc(sizeof(ModPlug));
	if (!file->sf) {
		delete mod.mod;
		free(data);
		return jfauderr_outofmemory;
	}

	sysf_close(&file->fh);	// we don't need to keep the file open

	memcpy(file->sf, &mod, sizeof(ModPlug));

	return jfauderr_noerror;
}

static jfauderr unload(File *file, char *noclose)
{
	ModPlug *mod = (ModPlug*)file->sf;
	
	if (mod->mod) {
		delete mod->mod;
	}
	// caller cleans up file->sf
	
	*noclose = 1;
	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file, JFAudMetaInfo what, char *buf, int *len)
{
	return jfauderr_notimpl;
}


// WaveformFileReader functions
static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	ModPlug *mod = (ModPlug *)file->sf;
	int br, nb;
	jfauderr err;

	nb = 2 * 2 * n;	// 2bytes * 2channels

	err = pcmbuffer_allocate(buf, n, min(conf.samplerate,44100), 2, 16);
	if (err != jfauderr_noerror) return err;

	mod->mod->SetRepeatCount(loop ? -1:0);
	buf->numsamples = mod->mod->Read(buf->data, nb);

	return jfauderr_noerror;
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	// not needed
	return jfauderr_notimpl;
}

static jfauderr getplaytime(File *file, float *len)
{
	*len = -1.0;	// always stream modules
	
	return jfauderr_noerror;
}

	
WaveformFileReader wf_mod = {
	"MOD/XM/IT/S3M/etc",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_mod,
	readsamples,
	readallsamples,
	getplaytime
};
	
}
